@@ -1,9 +1,11 @@
-from typing import Any, Awaitable, Callable, Collection, Deque, Dict, Iterable, List, Optional, Type
+from typing import Any, Callable, Collection, Deque, Dict, Iterable, List, Optional, Type
 from os.path import join
 from decimal import Decimal
 import json
 from datetime import datetime, timedelta, timezone
+# from time import time
 from collections import deque
+# from functools import wraps
 from types import TracebackType
 from asyncio import Condition, sleep
 
@@ -88,8 +90,8 @@ class Throttler:
     async def __aexit__(self,
                         exc_type: Optional[Type[BaseException]],
                         exc_value: Optional[BaseException],
-                        traceback: Optional[TracebackType]) -> Awaitable[Optional[bool]]:
-        pass
+                        traceback: Optional[TracebackType]) -> bool:
+        return False
 
     async def acquire(self) -> None:
         now: datetime = datetime.now(timezone.utc)
@@ -119,6 +121,28 @@ class Throttler:
         return datetime.now(timezone.utc)
 
 
+# def ratelimit(limit: int, every: float):
+
+#     def limitdecorator(func):
+#         logs: deque = deque()
+
+#         @wraps(func)
+#         async def new_func(*args, **kwargs):
+#             if len(logs) >= limit:
+#                 # before: time = logs.pop()
+#                 # now: time = time()
+#                 sleep_duration: float = every - (time() - logs.pop())
+#                 if sleep_duration > 0:
+#                     await sleep(sleep_duration)
+
+#             logs.appendleft(time())
+#             return await func(*args, **kwargs)
+
+#         return new_func
+
+#     return limitdecorator
+
+
 # class RateLimitQuota:
 
 #     def __init__(self, header: str = HEADER_RATE_LIMIT_REMAINING) -> None:
@@ -150,7 +174,7 @@ def is_missing_status(status: List = [404, 429]) -> Callable:
     return is_status
 
 
-def raise_for_status(response: ClientResponse, full_url: str = None, ignore_status: Collection = []) -> None:
+def raise_for_status(response: ClientResponse, full_url: Optional[str] = None, ignore_status: Collection = []) -> None:
 
     try:
         response.raise_for_status()
@@ -165,19 +189,32 @@ def raise_for_status(response: ClientResponse, full_url: str = None, ignore_stat
         # raise APIError(str(error)) from None
 
 
+def _retry_pattern() -> Callable:
+    return backoff.on_exception(
+        backoff.expo,
+        (json.decoder.JSONDecodeError, ClientResponseError, ClientError),
+        factor=3,
+        giveup=is_missing_status([404, 429]),
+        max_tries=5,
+        logger=LOGGER
+    )
+
+
 class Client:
 
     def __init__(self, config: Dict) -> None:
+        self.config: Dict = config
         self.url: str = config.get('url', '')
-        self.headers = {'Accept': 'application/json'} | config.get('headers', {})
-        self._auth = BasicAuth(config.get('login'), config.get('password', '')) if config.get('login') else None
+        self.headers: Dict[str, Any] = {'Accept': 'application/json'} | config.get('headers', {})
+        self._auth: Optional[BasicAuth] = BasicAuth(config.get('login'), config.get('password', '')) if config.get('login') else None
 
         # NOTE: Allowed 300 requests per minute (5 requests per second), with occasional bursts of up to 20 requests at a time.
         self._throttler: Throttler = Throttler(config.get('rate_limit', 5), config.get('rate_period', 1.0))
+        self.connect()
         LOGGER.debug('API Client: Throttler %s', json.dumps({'rate_limit': self._throttler._rate_limit, 'rate_period': self._throttler._period}))
 
     async def __aenter__(self) -> Any:
-        return self.connect()
+        return self
 
     async def __aexit__(
         self,
@@ -188,6 +225,8 @@ class Client:
         await self.close()
 
     def connect(self) -> Any:
+        # timeout_seconds: int = self.config.get('login', 99)
+        # session_timeout: ClientTimeout = ClientTimeout(total=None, sock_connect=timeout_seconds, sock_read=timeout_seconds)
         self.session = ClientSession(
             # NOTE: overwrite the default connector to customise the default connection settings applied to the queries
             connector=TCPConnector(
@@ -199,6 +238,8 @@ class Client:
                 # keepalive_timeout=30
             ),
             auth=self._auth,
+            headers=self.headers,
+            # timeout=session_timeout,
             connector_owner=True)
         LOGGER.debug('API Client: %s', json.dumps({'message': 'API connected'}))
         return self
@@ -206,33 +247,50 @@ class Client:
     async def close(self) -> None:
         await self.session.close()
 
-    @backoff.on_exception(
-        backoff.expo,
-        (json.decoder.JSONDecodeError, ClientResponseError, ClientError),
-        factor=3,
-        giveup=is_missing_status([404, 429]),
-        max_tries=5,
-        logger=LOGGER
-    )  # type: ignore
+    @_retry_pattern()
     # @backoff.on_exception(RateLimitQuota, ClientResponseError, jitter=None, max_tries=5, logger=LOGGER)  # type: ignore
-    async def get(self, *args: Optional[Any], **kwargs: Dict) -> Dict:
+    async def get(self, *args: Optional[Any], **kwargs: Dict) -> Iterable:
         path: str = str(args[0])
         full_url: str = join(self.url, str(kwargs.pop('endpoint', path)))
         LOGGER.debug('API Client GET: %s', json.dumps({'message': 'HTTP Request', 'url': full_url}))
 
-        async with self._throttler, self.session.request('get', full_url, headers=self.headers, params=kwargs.pop('filters', {})) as response:
+        async with self._throttler, self.session.get(full_url, params=kwargs.pop('params', {})) as response:
             with http_request_timer(path) as timer:
                 if response.status != 200:
                     raise_for_status(response, full_url=full_url, ignore_status=kwargs['ignore_status'] if 'ignore_status' in kwargs else [])
                     timer.tags[metrics.Tag.http_status_code] = response.status
-                    return {}
                 else:
                     return json.loads(await response.text(), parse_float=Decimal)
                     # NOTE: Alternative return await response.json(loads=partial(json.loads, parse_float=Decimal))
+        return {}
+
+    @_retry_pattern()
+    async def post(self, *args: Optional[Any], **kwargs: Dict) -> Iterable:
+        path: str = str(args[0])
+        full_url: str = join(self.url, str(kwargs.pop('endpoint', path)))
+        LOGGER.debug('API Client POST: %s', json.dumps({'message': 'HTTP Request', 'url': full_url}))
+
+        async with self._throttler, self.session.post(
+            full_url,
+            params=kwargs.pop('params', {}),
+            # headers=self.headers,
+            # timeout=9,
+            data=json.dumps(kwargs.pop('data', {}))
+        ) as response:
+            with http_request_timer(path) as timer:
+                if response.status != 200:
+                    raise_for_status(response, full_url=full_url, ignore_status=kwargs['ignore_status'] if 'ignore_status' in kwargs else [])
+                    timer.tags[metrics.Tag.http_status_code] = response.status
+                else:
+                    return json.loads(await response.text(), parse_float=Decimal)
+                    # NOTE: Alternative return await response.json(loads=partial(json.loads, parse_float=Decimal))
+        return {}
 
     async def get_records(self, tap_stream_id: str, format_response: Callable, *args: Optional[Any], **kwargs: Dict) -> Iterable:
 
-        response = await self.get(tap_stream_id, *args, **kwargs),
-        records = format_response(response) if callable(format_response) else []
+        def response_list(r: Any) -> tuple:
+            return (r,) if isinstance(r, dict) else r
 
-        return records
+        response = response_list(await self.get(tap_stream_id, *args, **kwargs))
+
+        return format_response(response) if callable(format_response) else response
